<?php
/**
 * BlueskyPost MediaWiki extension  version 0.2.0.1
 *	for details please see: https://www.mediawiki.org/wiki/Extension:BlueskyPost
 *
 * Copyright (c) 2023 Kimagurenote https://kimagurenote.net/
 * License: Revised BSD license http://opensource.org/licenses/BSD-3-Clause
 *
 * Function:
 *	Mediawiki extension to post update summary to Bluesky.
 *
 * Dependency:
 *	MediaWiki 1.35+ - use hook onPageSaveComplete(), PageSaveComplete replaced old PageContentInsertComplete and PageContentSaveComplete hooks since 1.35.
 *	https://www.mediawiki.org/wiki/Manual:Hooks/PageSaveComplete
 *	PHP 7.2.0+ - call json_encode() with JSON_INVALID_UTF8_IGNORE
 *	https://www.php.net/manual/ja/json.constants.php
 *
 * History:
 * 2023.09.21 Version 0.2.0.1
 *	ltrim description of link card
 *	(testing) max length with shorten url
 * 2023.09.20 Version 0.2.0
 *	suport to embed a card with image and description
 *	(testing) accesskey support
 *	https://www.mediawiki.org/wiki/Manual:Access_keys
 * 2023.09.13 Version 0.1.0
 *	1st test (support text and uri only)
 *
 * @file
 * @ingroup Extensions
 * @author Kimagurenote
 * @copyright © 2023 Kimagurenote
 * @license The BSD 3-Clause License
 */

if( !defined( 'MEDIAWIKI' ) ) {
	echo( "This file is an extension to the MediaWiki software and cannot be used standalone.\n" );
	die( 1 );
}

use MediaWiki\MediaWikiServices;
use HtmlFormatter\HtmlFormatter;

define( 'BLUESKYPOST_ID', 'wpBlueskyPost' );	// checkbox id and name
define( 'BLUESKYPOST_MAXDESCSIZE', 500 );	// max description size for card
define( 'BLUESKYPOST_MAXIMAGESIZE', 1000000 );	// max image size to upload
//define( 'BLUESKYPOST_DEBUG', 'blueskypost' );	// for debug

class BlueskyPost {
	/* Extension information */
	static $logger = null;
	/* AT Protocol parameters */
	static $atprotoMaxLen = 300;
	static $atprotoShrink = '…';
	static $atprotoHost = "bsky.social";
	static $atprotoSession = [];

	/**
	 * https://atproto.com/blog/create-post
	 */
	public static function atprotoCall( $endpoint, $data, $session = null ) {
		global $wgBlueskyPostTimeout;

		/* parameter check */
		$host = self::$atprotoHost;
		if( empty( $host ) or empty( $endpoint ) or empty( $data ) ) {
			return "";
		}

		/* make parameter to call API */
		$json = json_encode( $data, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_INVALID_UTF8_IGNORE );
		if( empty( $json ) ) {
			return "";
		}

		$param = [
			'http' => [
				'method' => "POST",
				'header' => [
					"Content-Type: application/json"
				],
				'content' => $json,
				'timeout' => (float)( $wgBlueskyPostTimeout ?? 10 )
			]
		];
		if( $session['accessJwt'] ) {
			$param['http']['header'][] = "Authorization: Bearer " . $session['accessJwt'];
		}
		$stream = stream_context_create( $param );

		/* call API */
		$ret = file_get_contents( "https://$host/xrpc/$endpoint", false, $stream );
		if( empty( $ret ) ) {
			return "";
		}

		return json_decode( $ret, true ) ?? "";
	}


	/**
	 * https://atproto.com/blog/create-post
	 */
	public static function atprotoLogin() {
		global $wgBlueskyPostIdentifier, $wgBlueskyPostPassword;

		/* parameter check */
		if( empty( $wgBlueskyPostIdentifier ) or empty( $wgBlueskyPostPassword ) ) {
			return "";
		}

		/* make parameter to call API */
		$data = [
			'identifier' => $wgBlueskyPostIdentifier,
			'password' => $wgBlueskyPostPassword
		];

		return self::atprotoCall( "com.atproto.server.createSession", $data ) ?? "";
	}


	/**
	 * https://atproto.com/blog/create-post#mentions-and-links
	 */
	public static function atprotoRichtextUrl( string &$text ) {

		$facets = [];
		$offset = 0;
		$maxlen = self::$atprotoMaxLen;
		while( ( $offset < strlen( $text ) ) and ( $offset < $maxlen ) and preg_match('/(https?:\/\/)[\S]+\.[\S]+/', $text, $matches, PREG_OFFSET_CAPTURE, $offset ) ) {
//			$maxlen += strlen( $matches[1][0] );
			$end = $matches[0][1] + strlen( $matches[0][0] );
			$s = mb_strcut( $text, 0, $end );	/* support multibyte chars */
			if( mb_strlen( $s ) > $maxlen ) {
				$end = strlen( mb_substr( $text, 0, $maxlen - mb_strlen( self::$atprotoShrink ) ) . self::$atprotoShrink );
			}
			$facets[] = [
				'index' => [
					'byteStart' => $matches[0][1],
					'byteEnd' => $end
				],
				'features' => [
					[
						'$type' => "app.bsky.richtext.facet#link",
						'uri' => $matches[0][0]
					]
				]
			];
			$offset = $end + 1;
		}

		/* shrink if $text over 300 chars (mutibyte chars support) */
		if( mb_strlen( $text ) > $maxlen ) {
			$text = mb_substr( $text, 0, $maxlen - mb_strlen( self::$atprotoShrink ) ) . self::$atprotoShrink;
		}

		//defined( 'BLUESKYPOST_DEBUG' ) && ( wfDebugLog( BLUESKYPOST_DEBUG, __FUNCTION__ . "() out: maxlen=$maxlen, text=" . $text ) );	// for debug

		return $facets;
	}


	/**
	 * https://atproto.com/blog/create-post
	 */
	public static function atprotoPost( $text, $langs = null, $createdAt = "", $embed = null ) {
		global $wgBlueskyPostLangs;

		/* create session (login) */
		if( empty( self::$atprotoSession ) ) {
			self::$atprotoSession = $session = self::atprotoLogin();
			if( empty( $session ) ) {
				return false;
			}
		} else {
			$session = self::$atprotoSession;
		}

		/* generate post data */
		if( empty( $createdAt ) ) {
			$createdAt = gmdate( "Y-m-d\TH:i:s" ) . ".00000Z";	// GMT
//			$createdAt = date_format( date_create(), "Y-m-d\TH:i:sP" );	// localtime
		}
		$facets = self::atprotoRichtextUrl( $text );
		$data = [
			'repo' => $session['did'],
			'collection' => "app.bsky.feed.post",
			'record' => [
				'$type' => "app.bsky.feed.post",
				'text' => $text,
				'createdAt' => $createdAt
			]
		];
		if( !empty( $facets ) ) {
			$data['record']['facets'] = $facets;
		}
		if( !empty( $langs ) and is_array( $langs ) ) {
			$data['record']['langs'] = $langs;
		} elseif( !empty( $wgBlueskyPostLangs ) and is_array( $wgBlueskyPostLangs ) ) {
			$data['record']['langs'] = $wgBlueskyPostLangs;
		}
		if( !empty( $embed ) ) {
			$data['record']['embed'] = $embed;
		}

		/* call API */
		return self::atprotoCall( "com.atproto.repo.createRecord", $data, $session );
	}


	/**
	 * https://atproto.com/blog/create-post#images-embeds
	 * https://atproto.com/specs/xrpc#blob-upload-and-download
	 * $img: array
	 *	$img['path']
	 *	$img['width']
	 *	$img['height']
	 * return $img
	 *	
	 */
	private static function atprotoUploadImage( $img ) {
		global $wgBlueskyPostTimeout;

		/* parameter check */
		if( !is_array( $img ) or !file_exists( $img['path'] ) ) {
			return null;
		}

		$host = self::$atprotoHost;
		if( empty( $host ) ) {
			return null;
		}

		/* get file size */
		$img['size'] = filesize( $img['path'] );
		if( $img['size'] > BLUESKYPOST_MAXIMAGESIZE ) {
			return null;
		}

		/* get mime type */
		$img['mime'] = mime_content_type( $img['path'] );
		if( !$img['mime'] ) {
			return null;
		}

		/* read image file */
		$data = file_get_contents( $img['path'] );
		if( !$data ) {
			return null;
		}

		/* create session (login) */
		if( empty( self::$atprotoSession ) ) {
			self::$atprotoSession = $session = self::atprotoLogin();
			if( empty( $session ) ) {
				return null;
			}
		} else {
			$session = self::$atprotoSession;
		}

		/* make parameter to call API */
		$param = [
			'http' => [
				'method' => "POST",
				'header' => [
					"Content-Type: " . $img['mime'],
					"Authorization: Bearer " . $session['accessJwt']
				],
				'content' => $data,
				'timeout' => (float)( $wgBlueskyPostTimeout ?? 10 )
			]
		];
		$stream = stream_context_create( $param );

		/* call API */
		$ret = file_get_contents( "https://$host/xrpc/com.atproto.repo.uploadBlob", false, $stream );
		if( empty( $ret ) ) {
			return null;
		}

		/* update $img and return */
		$json = json_decode( $ret, true );
		if( empty( $json['blob']['ref']['$link'] ) ) {
			return null;
		}
		$img['link'] = $json['blob']['ref']['$link'];
		if( !empty( $json['blob']['mimeType'] ) ) {
			$img['mime'] = $json['blob']['mimeType'];
		}
		if( !empty( $json['blob']['size'] ) ) {
			$img['size'] = $json['blob']['size'];
		}
		return $img;
	}


	/**
	 * https://atproto.com/blog/create-post#website-card-embeds
	 */
	public static function getCardFromTitle( $title, $desc = "", $thumb = null ) {
		global $wgSitename;

		/* Title exist?
		 * https://www.mediawiki.org/wiki/Manual:Title.php/ja
		 * https://doc.wikimedia.org/mediawiki-core/master/php/classMediaWiki_1_1Title_1_1Title.html
		 */
		if( empty( $title ) or !$title->exists() ) {
			return null;
		}

		/* make a card
		 * https://atproto.com/blog/create-post#website-card-embeds
		 */
		$embed = [
			'$type' => "app.bsky.embed.external",
			'external' => [
				'uri' => $title->getFullURL(),
				'title' => $title->getText(),
				'description' => $desc,
			]
		];
		if( !empty( $wgSitename ) ) {
			$embed['external']['title'] .= ' - ' . $wgSitename;
		}

		if( is_array( $thumb ) and !empty( $thumb['link'] ) ) {
			$embed['external']['thumb'] = [
				'$type' => "blob",
				'ref' => [
					'$link' => $thumb['link']
				],
				'mimeType' => $thumb['mime'],
				'size' => $thumb['size']
			];
		}

		return $embed;
	}


	/**
	 * get image thumbnail if Extension:PageImages installed
	 */
	private static function getImage( $title ) {

		/* Title exist?
		 * https://www.mediawiki.org/wiki/Manual:Title.php/ja
		 * https://doc.wikimedia.org/mediawiki-core/master/php/classMediaWiki_1_1Title_1_1Title.html
		 */
		if( empty( $title ) or !$title->exists() ) {
			return null;
		}

		/* Extension:PageImages installed?
		 * https://www.mediawiki.org/wiki/Manual:Extension_registration/Architecture
		 */
		if( !ExtensionRegistry::getInstance()->isLoaded( 'PageImages' ) ) {
			return null;
		}

		/* get page image */
		$file = \PageImages\PageImages::getPageImage( $title );
		if( empty( $file ) or !$file->exists() ) {
			return null;
		}

		/* get thumbnail */
		$thumbnail = $file->getThumbnailSource( [ 'physicalWidth' => 300 ] );
		if( $thumbnail ) {
			$thumbnail['alt'] = $title->getText();
		}

		return $thumbnail;
	}


	/**
	 * get a page description
	 */
	private static function getDescription( $wikiPage, $maxlen = BLUESKYPOST_MAXDESCSIZE ) {
		global $wgOut;

		/* parameter check */
		if( empty( $wikiPage ) ) {
			return "";
		}

		/* get description text
		 * https://doc.wikimedia.org/mediawiki-core/master/php/classParserOutput.html#a8232c6c65957276256d72040440147a6
		 */
//		$desc = $wikiPage->getParserOutput()->getText();	// parsed HTML

		/* get description text from wikitext
		 * https://doc.wikimedia.org/mediawiki-core/master/php/classWikiPage.html#a0de8dca25695458c15b2f8bfe4db0b2f
		 * https://doc.wikimedia.org/mediawiki-core/master/php/interfaceContent.html#a356bec57fd5c8edf3e7077dd44fd8789
		 * https://www.mediawiki.org/wiki/Manual:Page_content_models
		 * https://www.mediawiki.org/wiki/Manual:Parser.php
		 * https://doc.wikimedia.org/mediawiki-core/master/php/classParser.html
		 */
		$text = $wikiPage->getContent()->getText();	// wikitext
		$parser = MediaWikiServices::getInstance()->getParserFactory()->create();
		$desc = $parser->getSection( $text, 0 );

		if( empty( $desc ) ) {
			return "";
		}

		/* remove image and template */
		//$desc = preg_replace( '/\[\[(Image|File):.+?\]\]\n*/i', '', $desc );	// file (image)
		$desc = preg_replace( '/\{\{.+?\|.+?\}\}\n*/', '', $desc );	// template
		/* parse */
		$desc = $wgOut->parseAsContent( $desc );

		/* flatten HTML tags by HtmlFormatter
		 * https://www.mediawiki.org/wiki/HtmlFormatter
		 */
		$formatter = new HtmlFormatter( $desc );
		$formatter->remove( 'img' );
		$formatter->remove( 'table' );
		$formatter->flattenAllTags();
		$formatter->filterContent();
		$desc = $formatter->getText();

		/* replace crlf to lf */
		$desc = str_replace( "\r\n", "\n", $desc );	// Windows
		$desc = str_replace( "\r", "\n", $desc );	// MacOS
		/* replace lf to space */
		$desc = str_replace( "\n", " ", $desc );
		/* ltrim */
		$desc = ltrim( $desc, " ]}" );

		/* shorten */
		if( $maxlen > mb_strlen( self::$atprotoShrink ) and mb_strlen( $desc ) > $maxlen ) {
			$desc = mb_substr( $desc, 0, $maxlen - mb_strlen( self::$atprotoShrink ) ) . self::$atprotoShrink;
		}

		return $desc;
	}


	/**
	 * https://www.mediawiki.org/wiki/Manual:Hooks/EditPageGetCheckboxesDefinition
	 */
	public static function onEditPageGetCheckboxesDefinition( EditPage $editPage, array &$checkboxes ) {
		global $wgContentNamespaces, $wgBlueskyPostSuppressNew, $wgBlueskyPostSuppressEdit, $wgBlueskyPostChecked;

		/* check namespace */
		$title = $editPage->getTitle();
		$ns = $title->getNamespace();
		if( empty( $wgContentNamespaces ) ) {
			if( $ns != NS_MAIN ) {
				return true;
			}
		} elseif ( !in_array( $ns, $wgContentNamespaces, true ) ) {
			return true;
		}

		/* set checkbox parameters */
		/* https://doc.wikimedia.org/mediawiki-core/master/php/classMediaWiki_1_1EditPage_1_1EditPage.html#a90e727e84309de45ed551b621a04e617 */
		$check = [
			'label-message' => "",
			'id' => BLUESKYPOST_ID,
			'default' => ( $wgBlueskyPostChecked ? true : false ),
			'title-message' => "",
			'legacy-name' => BLUESKYPOST_ID
		];

		/* edit or new */
		if( $title->exists() ) {	/* edit page */
			if( $wgBlueskyPostSuppressEdit ) {
				return true;
			}
			$check['label-message'] = 'blueskypost-checknew' ?? "";
			$check['title-message'] = 'tooltip-blueskypost-checknew' ?? "";
		} else {	/* new page */
			if( $wgBlueskyPostSuppressNew ) {
				return true;
			}
			$check['label-message'] = 'blueskypost-checkedit' ?? "";
			$check['title-message'] = 'tooltip-blueskypost-checkedit' ?? "";
		}

		$checkboxes[BLUESKYPOST_ID] = $check;
		return true;
	}


	/**
	 * https://www.mediawiki.org/wiki/Manual:Hooks/PageSaveComplete
	 */
	public static function onPageSaveComplete( $wikiPage, $user, string $summary, int $flags, $revisionRecord, $editResult ) {
		global $wgRequest;

		/* summary empty */
		if( empty( $summary ) ) {
			return;
		}

		/* skip minor edit */
		if( $revisionRecord->isMinor() ) {
			return;
		}

		/* check checkbox */
		if( empty( $wgRequest->getCheck( BLUESKYPOST_ID ) ) ) {
			return;
		}

		if( $flags & EDIT_NEW ) {
			$key = 'blueskypost-post-new';
		} elseif( $flags & EDIT_UPDATE ) {
			$key = 'blueskypost-post-edit';
		} else {
			return;
		}

		/* generate text to post */
		$title = $wikiPage->getTitle();
		$basetext = $title->getBaseText();
		$fullurl = $title->getFullURL();
		$text = wfMessage( $key, $basetext, $fullurl, $summary )->inContentLanguage()->plain() ?? $summary;

		/* generate link card */
		/* get image thumbnail if Extension:PageImages installed */
		$thumbnail = self::getImage( $title );
		/* upload thumbnail to embed card */
		$thumbnail = self::atprotoUploadImage( $thumbnail );
		/* get card from title and thumbnail */
		$card = self::getCardFromTitle( $title, self::getDescription( $wikiPage ), $thumbnail ?? null );

		/* generate timecode for createdAt */
		$timestamp = $revisionRecord->getTimestamp();
		if( $timestamp and strlen( $timestamp ) === 14 ) {
			$time = sscanf( $timestamp, "%4s%2s%2s%2s%2s%2s" );
			$date = sprintf( "%4d-%02d-%02dT%02d:%02d:%02d.000000Z", $time[0], $time[1], $time[2], $time[3], $time[4], $time[5] );
		} else {
			$date = "";
		}

		/* language code if provided *
		//$lang = $wikiPage->getLanguage();
		$lang = $wikiPage->getParserOutput()->getLanguage();
		$lang = empty( $lang ) ? null : [ $lang ];
		 * not work on MediaWiki 1.35 */
		$lang = null;

		/* post to Bluesky (call AT Protocol) */
		self::atprotoPost( $text, $lang, $date, $card ?? null );

		return;
	}
}
